"""
data_generator.py

Script to generate pseudo-random artificial data for games.
Uses distribution curves to place sprites and create bounding boxes.

@authors: Jaden Kim & Chanha Kim
"""

# modules
import os
import glob
import shutil
import argparse
import yaml
from PIL import Image
from numpy.random import choice
from yards.tools.helper import * # import the helper module correctly

# load config parameters
def load_config(config_file):
    """Returns a configuration file as a dictionary"""
    with open(r'{}'.format(config_file)) as file:
        all_params = yaml.load(file, Loader=yaml.FullLoader)

    return all_params


# parse all_params
def parse_params(all_params):
    """Parses all_params dictionary into separate dictionaries"""
    directories = all_params.pop('directories')
    params = all_params.pop('parameters')
    classes = all_params.pop('classes')
    class_numbers = {c: n for (c, n) in list(zip(classes, [i for i in range(len(classes))]))}
    if not params['label_all_classes']:
        for c in classes:
            if c not in params['labeled_classes']:
                class_numbers[c] = -1

    return directories, params, (classes, class_numbers)


# create output directories
def _create_output_dirs(output_dir):
    """Creates the output directories for dataset if they don't exist"""
    if os.path.isdir(output_dir):
        shutil.rmtree(output_dir)
        os.mkdir(output_dir)
    parent_out_dir = (output_dir+out for out in ['images/', 'labels/'])
    child_out_dir = (output_dir+out for out in ['images/train/', 'images/val/', 'labels/train/', 'labels/val/'])
    for out in parent_out_dir: os.mkdir(out)
    for out in child_out_dir: os.mkdir(out)

    return child_out_dir


# create an image
def _create_image(dirs, class_info, params, count, output_dir):
    """Creates a dataset image at the desired directory. Returns bbox cache for creating annotations."""
    # open the map and create a copy of it as a new image
    background_image = Image.open(choice(glob.glob(dirs['maps']+'.png')))
    new_image = background_image.copy().convert('RGBA')
    map_dim = new_image.size
    background_image.close()

    # get the sprite paths
    sprite_paths = _gather_sprite_paths(dirs['sprites'], class_info, params['random_distribution'], params['maximum_sprites_per_image'])
    bbox_cache = []

    # add the sprites to the new image, saving the bounding box information in a cache
    for sprite_path in sprite_paths:
        sprite = Image.open(sprite_path).convert('RGBA')
        if params['transform_sprites']:
            sprite = _transform_sprite(sprite)

        sprite_pos, sprite_clipped = _get_sprite_pos(map_dim, sprite.size, params['clip_sprites'])
        if sprite_clipped:
            sprite, sprite_dim, sprite_pos = _edge_handler(map_dim, sprite.size, sprite_pos, sprite)

        new_image = _draw_sprite_to_background(sprite, new_image, sprite_pos)
        sprite.close()
        bbox = _get_bbox(map_dim, sprite_dim, sprite_pos)
        if sprite_paths[sprite_path] != -1:
            bbox_cache.append((sprite_paths[sprite_path], *bbox))

    # save and close the image
    new_image.save('{}{}.png'.format(output_dir, count))
    new_image.close()

    # return the cache
    return bbox_cache


# create an annotation
def _create_annotation(bbox_cache, count, output_dir):
    """Creates a dataset label at the desired directory."""
    with open('{}{}.txt'.format(output_dir, count)) as file:
        for bbox in bbox_cache:
            file.write('{} {} {} {} {}\n'.format(*bbox))


# run a loop
def loop(all_params):
    """Creates the desired number of images and labels for the dataset."""
    # load parameters
    dirs, params, class_info = all_params
    train_size = int(params['num_images']*params['train_size'])
    image_train_dir, image_val_dir, label_train_dir, label_val_dir = _create_output_dirs(directories['output_dir'])

    # run loop
    print('Writing {} images.'.format(params['num_images']))
    for count in progressbar.progressbar(range(1, num_images+1)):
        bbox_cache = _create_image(dirs, class_info, params, count, image_train_dir if count <= train_size else image_val_dir)
        _create_annotation(bbox_cache, count, label_train_dir if count <= train_size else label_val_dir)
    print('Finished writing {} images.'.format(params['num_images']))

    
# visualize data
def visualize(output_dir, num_visualize):
    """Visualizes some or all of the images."""
    examples_path = output_dir + 'examples/'
    if os.path.isdir(examples_path):
        shutil.rmtree(examples_path)
        os.mkdir(examples_path)

    for image_path in glob.glob(output_dir+'.png'):



"""
yards is a class version of the original yards program.

Script to generate pseudo-random artificial data for games.
Uses distribution curves to place sprites and create bounding boxes.

@authors: Jaden Kim & Chanha Kim
@date   : 7/20/2020
"""

# modules
import os
import glob
import shutil
import yaml
import tqdm
from PIL import Image, ImageDraw
from numpy.random import choice
from numpy import unique
from .tools import _helper # import the helper module correctly
from .tools import _validator

class yards():

    def __init__(self, config_path = None):
        '''Initializing a yards object'''
        if config_path != None:
            self.load_config_from_file(config_path)
        else:
            self._config_path = None
            self._config = None
            self._dirs = None
            self._params = None
            self._classes = None
            self._class_numbers = None
            self._output_dirs = None
            self._map_path_cache = None
            self._sprite_path_cache = None

    # Setting configurations and getters/setters

    def _create_output_dirs(self):
        '''Helper that creates output directories'''
        if os.path.isdir(self._dirs['output']):
            shutil.rmtree(self._dirs['output'])
            os.mkdir(self._dirs['output'])

        os.mkdir(self._dirs['output'] +'images/')
        os.mkdir(self._dirs['output'] +'labels/')

        self._output_dirs = {
            'images_train': self._dirs['output'] + 'images/train/',
            'images_val': self._dirs['output'] + 'images/val/',
            'labels_train': self._dirs['output'] + 'labels/train/',
            'labels_val': self._dirs['output'] + 'labels/val/'
        }
        for out in self._output_dirs.values(): os.mkdir(out)

    def _parse_params(self):
        '''Parses all_params dictionary into separate dictionaries'''
        self.set_directories(self._config['directories'])
        self.set_parameters(self._config['parameters'])
        self.set_classes(self._config['classes'])

    def _cache_paths(self):
        '''Caches all the params into separate dictionaries'''
        self._map_path_cache = glob.glob(self._dirs['maps']+'*.png')
        self._sprite_path_cache = {}
        for c in self._classes:
            self._sprite_path_cache[c] = glob.glob(self._dirs['sprites']+'{}/*.png'.format(c))

    def load_config_from_file(self, config_path):
        '''Loads configuration from a file'''
        self._config_path = config_path
        with open(r'{}'.format(self._config_path)) as file:
            self._config = yaml.load(file, Loader=yaml.FullLoader)
        self._parse_params()
        self._create_output_dirs()

    def set_config(self, config):
        '''Sets a configuration from a dictionary'''
        if _validator.validate_config(config):
            self._config = config
            self._parse_params()
            self._create_output_dirs()
        else:
            print('Configuration is not valid')

    def get_config(self):
        '''Returns the configuration'''
        return self._config

    def set_directories(self, paths):
        '''Sets the directories'''
        if _validator.validate_directories(paths):
            self._dirs = paths
        else:
            print('Directories are not valid')

    def get_directories(self):
        '''Returns the directories'''
        return self._dirs

    def set_parameters(self, parameters):
        '''Sets the parameters'''
        if _validator.validate_parameters(parameters):
            self._params = parameters
            self._params['num_train'] = self._params['train_size'] * self._params['num_images']
        else:
            print('Parameters are not valid')

    def get_parameters(self):
        '''Returns the parameters'''
        return self._parameters

    def set_classes(self, classes):
        '''Sets the classes'''
        if _validator.validate_classes(classes, self._params['classification_scheme']):
            if self._params['classification_scheme'] == 'mimic-real':
                self.approximate_frequency_spaces_from_real_data(classes)
            else:
                self._classes = classes
            if self._params != None:
                if self._params['label_all_classes']:
                    self._class_numbers = {c: n for (c, n) in list(zip(self._classes, [i for i in range(len(self._classes))]))}
                else:
                    self._class_numbers = {c: n for (c, n) in list(zip(self._params['labeled_classes'], [i for i in range(len(self._params['labeled_classes']))]))}
                    for c in list(set(self._classes.keys()) - set(self._params['labeled_classes'])):
                        self._class_numbers[c] = -1
            else:
                print("You haven't loaded the parameters yet. If you're labeling select classes, then load parameters first before loading classes.")

            if self._dirs != None:
                self._cache_paths()
            else:
                print("You haven't loaded the directory paths yet. Load the directories first before loading the classes.")

        else:
            print('Classes are not valid.')

    def get_classes(self):
        '''Returns the classes'''
        return self._classes

    def approximate_frequency_spaces_from_real_data(self, class_ids):
        """Approximates frequency spaces by analyzing real samples."""
        # set up some local variables
        real_samples_labels_dir = self._dirs['real_samples'] + "labels/"
        real_samples_labels_paths = glob.glob(real_samples_labels_dir + "*.txt")
        num_samples = len(real_samples_labels_paths)
        classes = {num: [] for num in class_ids.values()}
        print("Approximating frequency spaces from {num} samples...".format(num = num_samples))
        # read in class info
        for path in tqdm.tqdm(real_samples_labels_paths):
            for c in classes: classes[c].append(0)
            with open(path, "r") as file:
                for line in file:
                    class_id = int([float(i) for i in line.split()][0])
                    if class_id in classes:
                        classes[class_id][-1] += 1

        # analyze frequencies
        for c in class_ids:
            class_id = class_ids[c]
            classes[class_id] = list(unique(classes[class_id], return_counts = True))
            total = sum(classes[class_id][1])
            classes[class_id][1] = [num / total for num in classes[class_id][1]]
            classes[class_id][0] = list(classes[class_id][0])
            classes[c] = classes[class_id]
            del classes[class_id]
            print(c + ":\t\t", classes[c])

        print("Finished approximating frequency spaces.")

        self._classes = classes

    def _create_image(self, count, output_dir):
        '''Creates an image'''
        background_image = Image.open(choice(self._map_path_cache))
        new_image = background_image.copy().convert('RGBA')
        map_dim = new_image.size
        background_image.close()

        # get the sprite paths
        sprite_paths = _helper.gather_sprite_paths(self._sprite_path_cache, (self._classes, self._class_numbers), self._params['classification_scheme'], self._params['max_sprites_per_class'])
        bbox_cache = []

        # add the sprites to the new image, saving the bounding box information in a cache
        for sprite_path in sprite_paths:
            sprite = Image.open(sprite_path).convert('RGBA')
            if self._params['transform_sprites']:
                sprite = _helper.transform_sprite(sprite, map_dim)
            sprite_dim = sprite.size

            sprite_pos, is_sprite_clipped = _helper.get_sprite_pos(map_dim, sprite_dim, self._params['clip_sprites'])
            if is_sprite_clipped:
                sprite, sprite_dim, sprite_pos = _helper.edge_handler(map_dim, sprite_dim, sprite_pos, sprite)

            new_image = _helper.draw_sprite_to_background(sprite, new_image, sprite_pos)
            sprite.close()
            bbox = _helper.get_bbox(map_dim, sprite_dim, sprite_pos)
            if sprite_paths[sprite_path] != -1:
                bbox_cache.append((sprite_paths[sprite_path], *bbox))

        # save and close the image
        new_image.save('{}{}-{}.png'.format(output_dir, self._params['game_title'], count))
        new_image.close()

        # return the cache
        return bbox_cache

    def _create_annotation(self, bbox_cache, count, output_dir):
        """Creates a dataset label at the desired directory."""
        with open('{}{}-{}.txt'.format(output_dir, self._params['game_title'], count), 'w') as file:
            for bbox in bbox_cache:
                file.write('{} {} {} {} {}\n'.format(*bbox))

    def _create_image_and_annotate(self, count):
        bbox_cache = self._create_image(count, self._output_dirs['images_train'] if count <= self._params['num_train'] else self._output_dirs['images_val'])
        self._create_annotation(bbox_cache, count, self._output_dirs['labels_train'] if count <= self._params['num_train'] else self._output_dirs['labels_val'])
        return None

    def loop(self):
        """Creates the images."""
        print('Writing {} images...'.format(self._params['num_images']))
        for count in tqdm.tqdm(range(1, 1+self._params['num_images'])):
            bbox_cache = self._create_image(count, self._output_dirs['images_train'] if count <= self._params['num_train'] else self._output_dirs['images_val'])
            self._create_annotation(bbox_cache, count, self._output_dirs['labels_train'] if count <= self._params['num_train'] else self._output_dirs['labels_val'])
        print('Finished writing {} images.'.format(self._params['num_images']))

    def visualize(self, directory='train', num_visualize=50):
        """Draws bounding boxes around the images."""
        if directory == 'train' or directory == 'val':
            if directory == 'train':
                image_dir = self._output_dirs['images_train']
                label_dir = self._output_dirs['labels_train']
            elif directory == 'label':
                image_dir = directory['images']
                label_dir = directory['labels']
            examples_dir = self._dirs['output']+'examples/'
            
        else:
            if directory[-1] != '/':
                directory += '/'
            image_dir = directory + 'images/'
            label_dir = directory + 'labels/'
            examples_dir = directory+'examples/'

        if os.path.isdir(examples_dir):
            shutil.rmtree(examples_dir)
        os.mkdir(examples_dir)

        image_paths = glob.glob(image_dir+'*.png')[0:num_visualize]

        print('Visualizing {} example images...'.format(num_visualize))
        colors = {}
        for i in tqdm.tqdm(range(len(image_paths))):
            bboxes = []
            image_name = os.path.splitext(os.path.split(image_paths[i])[1])[0]
            with open(label_dir+'{}.txt'.format(image_name)) as file:
                for line in file:
                    line = [float(i) for i in line.split()]
                    bboxes.append(line)
            image = Image.open(image_paths[i])
            draw = ImageDraw.Draw(image)
            for bbox in bboxes:
                c = int(bbox[0])
                x1 = int(image.size[0]*bbox[1] - image.size[0]*bbox[3]/2)
                y1 = int(image.size[1]*bbox[2] - image.size[1]*bbox[4]/2)
                x2 = int(image.size[0]*bbox[1] + image.size[0]*bbox[3]/2)
                y2 = int(image.size[1]*bbox[2] + image.size[1]*bbox[4]/2)
                if c not in list(colors.keys()):
                    colors[c] = tuple(choice(range(256), size=3))
                draw.rectangle([x1, y1, x2, y2], fill=None, outline=colors[c])

            image.save(examples_dir+'annotated-{}.png'.format(image_name))
            image.close()
        print('Finished visualizing {} example images.'.format(num_visualize))





"""
"validator modules that validates packages"
@author: Jaden Kim & Chanha Kim
@date  : 7/20/2020
"""

def validate_config(config):
    '''Returns true if the config is valid'''
    correct_keys = {'directories', 'parameters', 'classes'}
    config_keys = set(config.keys())

    return (correct_keys - config_keys) == set()


def validate_directories(paths):
    '''Returns true if the directory paths are valid'''
    from os.path import isdir
    from os import mkdir
    
    # ensure that the keys are correct
    correct_keys = {'maps', 'sprites', 'output'}    # real data is optional
    dirs_keys = set(paths.keys())
    are_keys_correct = min([key in dirs_keys for key in correct_keys])
    # ensure that the directories exist
    dirs_paths = list(paths.values())
    if not isdir(paths['output']):
        mkdir(paths['output'])

    are_paths_correct = min([isdir(path) for path in dirs_paths])

    return are_keys_correct and are_paths_correct


def validate_parameters(parameters):
    '''Returns true if the parameters are valid'''
    correct_keys = {'game_title', 'num_images', 'train_size', 'mix_size',
                   'label_all_classes', 'labeled_classes', 'max_sprites_per_class',
                   'transform_sprites', 'clip_sprites', 'classification_scheme'}
    params_keys = set(parameters.keys())
    are_keys_correct = (correct_keys - params_keys) == set()

    def checklist(ls):
        ret = min([isinstance(el, str) for el in ls])
        return ret

    are_values_correct = True
    if not isinstance(parameters['game_title'], str):
        are_values_correct = False
    if not isinstance(parameters['num_images'], int) or parameters['num_images'] < 0:
        are_values_correct = False
    if not isinstance(parameters['train_size'], float) or parameters['train_size'] < 0.0 or parameters['train_size'] > 1.0:
        are_values_correct = False
    if not isinstance(parameters['mix_size'], float) or parameters['mix_size'] < 0.0 or parameters['mix_size'] > 1.0:
        are_values_correct = False
    if not isinstance(parameters['label_all_classes'], bool):
        are_values_correct = False
    if not parameters['label_all_classes'] and not checklist(parameters['labeled_classes']):
        are_values_correct = False
    if not isinstance(parameters['max_sprites_per_class'], int):
        are_values_correct = False
    if not isinstance(parameters['transform_sprites'], bool):
        are_values_correct = False
    if not isinstance(parameters['clip_sprites'], bool):
        are_values_correct = False
    if not (parameters['classification_scheme'] == 'distribution' or parameters['classification_scheme'] == 'discrete' or parameters['classification_scheme'] == 'random' or parameters['classification_scheme'] == 'mimic-real'):
        are_values_correct = False

    return are_keys_correct and are_values_correct


def validate_classes(classes, classification_scheme):
    '''Returns true if the classes are valid'''
    classes_keys = list(classes.keys())
    are_keys_correct = min([type(key) == str for key in classes_keys])
    classes_values = list(classes.values())
    if classification_scheme == 'distribution':
        are_values_correct = min([type(value) == list for value in classes_values]) and min([all(isinstance(sub_value, float) for sub_value in value) for value in classes_values])
    elif classification_scheme == 'random':
        are_values_correct = min([type(value) == int for value in classes_values])
    elif classification_scheme == 'discrete':
        are_values_correct = (min(classes_values) == max(classes_values)) and min([type(value) == int for value in classes_values])
    elif classification_scheme == 'mimic-real':
        are_values_correct = min([type(value) == int for value in classes_values])
    else:
        are_values_correct = False

    return are_keys_correct and are_values_correct




"""
Helper Functions for Main Script

@authors: Jaden Kim & Chanha Kim
"""
import numpy as np
from numpy.random import choice, randint
from PIL import Image, ImageOps
import os, glob, shutil


def _get_sprite_counts(classes, classification_scheme='distribution', sprite_cap=-1):
    """Returns a <class, count> dictionary of sprite counts
        for each class to be pasted in the new image."""
    sprite_counts = {}
    # first handle the raw counts
    if classification_scheme == 'random':
        for c in classes:
            sprite_counts[c] = choice([i for i in range(classes[c]+1)])
            if sprite_counts[c] > sprite_cap and sprite_cap != -1:
                sprite_counts[c] = sprite_cap
    elif classification_scheme == 'distribution':
        for c in classes:
            sprite_counts[c] = choice([i for i in range(len(classes[c]))], p=classes[c])
            if sprite_counts[c] > sprite_cap and sprite_cap != -1:
                sprite_counts[c] = sprite_cap
    elif classification_scheme == 'discrete':
        class_list = list(classes.keys())
        num_players = classes[class_list[0]]
        for i in range(num_players):
            c = choice(class_list)
            if c in list(sprite_counts.keys()):
                sprite_counts[c] += 1
            else:
                sprite_counts[c] = 1
    elif classification_scheme == 'mimic-real':
        for c in classes:
            sprite_counts[c] = choice(classes[c][0], p=classes[c][1])
            if sprite_counts[c] > sprite_cap and sprite_cap != -1:
                sprite_counts[c] = sprite_cap
    else:
        print('Invalid classification scheme.')

    return sprite_counts


def _get_sprite_path(class_path_cache):
    """Returns the file path of a random sprite from the sprite directory."""
    return choice(class_path_cache)


def gather_sprite_paths(sprite_path_cache, class_info, classification_scheme, sprite_cap):
    """Returns a <path, class number> dictionary of the file paths of all the sprites to be pasted on an image."""
    classes, class_numbers = class_info
    sprite_paths = {}
    sprite_counts = _get_sprite_counts(classes, classification_scheme, sprite_cap)
    for c in sprite_counts:
        for i in range(sprite_counts[c]):
            sprite_paths[_get_sprite_path(sprite_path_cache[c])] = class_numbers[c]

    return sprite_paths


def transform_sprite(sprite, map_dim):
    transformed_sprite = sprite
    operations = {"mirror":bool(choice(2)), "rotate":choice([0, Image.ROTATE_90, Image.ROTATE_180, Image.ROTATE_270]), "resize":bool(choice(2))}

    if operations["mirror"]:
        transformed_sprite = transformed_sprite.transpose(Image.FLIP_LEFT_RIGHT)

    if operations["rotate"] != 0:
        transformed_sprite = transformed_sprite.transpose(operations["rotate"])

    if operations["resize"] and (transformed_sprite.size[0] < int(map_dim[0]/2) and transformed_sprite.size[1] < int(map_dim[1]/2)):
        transformed_sprite = transformed_sprite.resize((transformed_sprite.size[0]*2, transformed_sprite.size[1]*2), Image.NEAREST)

    return transformed_sprite


def get_sprite_pos(map_dim, sprite_dim, clip_sprites):
    """Returns a sprite position for a given map and sprite."""
    map_w, map_h = map_dim
    sprite_w, sprite_h = sprite_dim

    if clip_sprites:
        x_pos = randint(-sprite_w, map_w)
        y_pos = randint(-sprite_h, map_h)
    else:
        x_pos = randint(0, map_w - sprite_w)
        y_pos = randint(0, map_h - sprite_h)

    sprite_clipped = True if ((x_pos < 0 or x_pos >= map_w - sprite_w) or (y_pos < 0 or y_pos >= map_h - sprite_h)) else False

    return (x_pos, y_pos), sprite_clipped


def _get_transparencies(sprite):
    """Returns the 'transparent' quadrants for an image."""
    # get a binarray array from the image
    bin_array = (np.asarray(sprite.convert('RGBA').getchannel('A')) > 128).astype(int)

    # find indices
    indices = np.argwhere(bin_array == 1)
    min_row, min_col = np.amin(indices, axis=0 )
    max_row, max_col = np.amax(indices, axis=0)

    # find row/col x min/max values
    min_row_cols = indices[np.squeeze(np.argwhere(indices[:, 0] == min_row)), :].reshape(-1, 2)
    min_col_rows = indices[np.squeeze(np.argwhere(indices[:, 1] == min_col)), :].reshape(-1, 2)
    max_row_cols = indices[np.squeeze(np.argwhere(indices[:, 0] == max_row)), :].reshape(-1, 2)
    max_col_rows = indices[np.squeeze(np.argwhere(indices[:, 1] == max_col)), :].reshape(-1, 2)

    # (x,y) positions of the quadrants
    ul = {'x':np.min(min_row_cols[:, 1]), 'y':np.min(min_col_rows[:, 0])}
    ur = {'x':np.max(min_row_cols[:, 1]), 'y':np.min(max_col_rows[:, 0])}
    bl = {'x':np.min(max_row_cols[:, 1]), 'y':np.max(min_col_rows[:, 0])}
    br = {'x':np.max(max_row_cols[:, 1]), 'y':np.max(max_col_rows[:, 0])}

    return (ul, ur, bl, br)


def edge_handler(map_dim, sprite_dim, sprite_pos, sprite):
    """Handles edge cases"""
    # load parameters
    (map_w, map_h), (sprite_w, sprite_h), (x_pos, y_pos) = map_dim, sprite_dim, sprite_pos
    ul, ur, bl, br = _get_transparencies(sprite)

    # left edge
    if x_pos < 0:
        if np.max([sprite_w - ur['x'], sprite_w - br['x']]) >= (x_pos + sprite_w):
            crop_x = int(np.average([ur['x'], br['x'], ul['x'], bl['x']]))
        else:
            crop_x = abs(x_pos)
        sprite = sprite.crop((crop_x, 0, sprite_w, sprite_h))
        sprite_w = sprite.size[0]
        x_pos = 0
    # right edge
    elif x_pos >= map_w - sprite_w:
        if np.max([ul['x'], bl['x']]) >= (map_w - x_pos):
            crop_x = int(np.average([ul['x'], bl['x'], ur['x'], br['x']]))
        else:
            crop_x = map_w - x_pos
        sprite = sprite.crop((0, 0, crop_x, sprite_h))
        sprite_w = sprite.size[0]
        x_pos = map_w - sprite_w

    # top edge
    if y_pos < 0:
        if np.max([sprite_h - bl['y'], sprite_h - br['y']]) >= (y_pos + sprite_h):
            crop_y = int(np.average([bl['y'], br['y'], ul['y'], ur['y']]))
        else:
            crop_y = abs(y_pos)
        sprite = sprite.crop((0, crop_y, sprite_w, sprite_h))
        sprite_h = sprite.size[1]
        y_pos = 0
    # bottom edge
    elif y_pos >= map_h - sprite_h:
        if np.max([ul['y'], ur['y']]) >= (map_h - y_pos):
            crop_y = int(np.average([ul['y'], ur['y'], bl['y'], br['y']]))
        else:
            crop_y = map_h - y_pos
        sprite = sprite.crop((0, 0, sprite_w, crop_y))
        sprite_h = sprite.size[1]
        y_pos = map_h - sprite_h

    return sprite, (sprite_w, sprite_h), (x_pos, y_pos)


def get_bbox(map_dim, sprite_dim, sprite_pos):
    """Returns the bounding box for a sprite that was pasted at position pos"""
    (map_w, map_h), (sprite_w, sprite_h), (x_pos, y_pos) = map_dim, sprite_dim, sprite_pos
    rel_x = (x_pos + sprite_w / 2) / map_w
    rel_y = (y_pos + sprite_h / 2) / map_h
    rel_w = sprite_w / map_w
    rel_h = sprite_h / map_h

    return (rel_x, rel_y, rel_w, rel_h)


def draw_sprite_to_background(sprite, background, pos):
    """Draws the sprite to background and returns the background."""
    background.alpha_composite(sprite, dest=pos)

    return background



"""
_cli.py is the command-line program for that the user can use to generate pseudo-random artificial data for games.

@author: Jaden Kim & Chanha Kim
"""

import argparse
import os
# from multiprocessing import cpu_count
from .yards import yards

# get arguments
def _get_args():
    """Parses and returns command-line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', '-c',
        type=str,
        default=None,
        help='The path to the config.yaml file.'
    )
    parser.add_argument('--visualize', '-v',
        nargs='*',
        default=None,
        help='Whether or not to visualize the output.'
    )
    # parser.add_argument('--parallel', '-p',
    #     action='store_true',
    #     help='Whether or not to parallelize the processes.'
    # )
    
    return parser.parse_args()

def _valid_config(config_path):
    '''Returns true if the path exists'''
    return config_path != None and os.path.exists(config_path)

def _valid_visualize(arguments):
    '''Returns true if arguments are valid'''
    not_none = arguments != None
    if not_none:
        if len(arguments) == 1:
            valid = arguments[0].isnumeric()
        elif len(arguments) == 2:
            valid = os.path.exists(arguments[0]) and arguments[1].isnumeric()
    else:
        valid = False
    return not_none and valid

def main():
    '''Entry point for cli interface'''
    args = _get_args()
    yd = yards()

    if _valid_config(args.config):
        yd.load_config_from_file(args.config)
        yd.loop()
        
        # if args.parallel:
        #     num_cpus = cpu_count()
        #     yd.parallel_loop(num_cpus=num_cpus)
        # else:
        #     yd.loop()

    if _valid_visualize(args.visualize):
        if len(args.visualize) == 1:
            yd.visualize(num_visualize=int(args.visualize[0]))
        elif len(args.visualize) == 2:
            yd.visualize(directory=args.visualize[0], num_visualize=int(args.visualize[1]))








# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yards', 'yards.tools']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.0,<2.0.0',
 'pillow>=7.2.0,<8.0.0',
 'progressbar2>=3.51.4,<4.0.0',
 'pyyaml>=5.3.1,<6.0.0',
 'tqdm>=4.48.0,<5.0.0']

entry_points = \
{'console_scripts': ['yards = yards._cli:main']}

setup_kwargs = {
    'name': 'yards',
    'version': '0.1.2',
    'description': '',
    'long_description': None,
    'author': 'chanhakim',
    'author_email': 'chanhakim17@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
